import { RawSchemaScope, hasArkKind } from "@arktype/schema";
import { domainOf, hasDomain, isThunk, throwParseError } from "@arktype/util";
import { Generic } from "./generic.js";
import { createMatchParser } from "./match.js";
import { parseObject, writeBadDefinitionTypeMessage } from "./parser/definition.js";
import { parseGenericParams } from "./parser/generic.js";
import { DynamicState } from "./parser/string/reduce/dynamic.js";
import { fullStringParse } from "./parser/string/string.js";
export const scope = ((def, config = {}) => new RawScope(def, config));
export class RawScope extends RawSchemaScope {
    parseCache = {};
    constructor(def, config) {
        const aliases = {};
        for (const k in def) {
            const parsedKey = parseScopeKey(k);
            aliases[parsedKey.name] =
                parsedKey.params.length ?
                    // TODO: this
                    new Generic(parsedKey.params, def[k], {})
                    : def[k];
        }
        super(aliases, config);
    }
    type(...args) {
        if (args.length === 1) {
            // treat as a simple definition
            return this.parseRoot(args[0]);
        }
        if (args.length === 2 &&
            typeof args[0] === "string" &&
            args[0][0] === "<" &&
            args[0].at(-1) === ">") {
            // if there are exactly two args, the first of which looks like <${string}>,
            // treat as a generic
            const params = parseGenericParams(args[0].slice(1, -1));
            const def = args[1];
            // TODO: validateUninstantiatedGeneric, remove this cast
            return new Generic(params, def, this);
        }
        // otherwise, treat as a tuple expression. technically, this also allows
        // non-expression tuple definitions to be parsed, but it's not a supported
        // part of the API as specified by the associated types
        return this.parseRoot(args);
    }
    match = createMatchParser(this);
    declare() {
        return { type: this.type.bind(this) };
    }
    define(def) {
        return def;
    }
    preparseRoot(def) {
        if (isThunk(def) && !hasArkKind(def, "generic")) {
            return def();
        }
        return def;
    }
    parseRoot(def, opts) {
        return this.parse(def, {
            args: { this: {} },
            $: this,
            ...opts
        }).bindScope(this);
    }
    parse(def, ctx) {
        if (typeof def === "string") {
            if (ctx.args && Object.keys(ctx.args).every((k) => !def.includes(k))) {
                // we can only rely on the cache if there are no contextual
                // resolutions like "this" or generic args
                return this.parseString(def, ctx);
            }
            if (!this.parseCache[def]) {
                this.parseCache[def] = this.parseString(def, ctx);
            }
            return this.parseCache[def];
        }
        return hasDomain(def, "object") ?
            parseObject(def, ctx)
            : throwParseError(writeBadDefinitionTypeMessage(domainOf(def)));
    }
    parseString(def, ctx) {
        return (this.maybeResolveNode(def) ??
            ((def.endsWith("[]") &&
                this.maybeResolveNode(def.slice(0, -2))?.array()) ||
                fullStringParse(new DynamicState(def, ctx))));
    }
}
export const writeShallowCycleErrorMessage = (name, seen) => `Alias '${name}' has a shallow resolution cycle: ${[...seen, name].join(":")}`;
export const parseScopeKey = (k) => {
    const firstParamIndex = k.indexOf("<");
    if (firstParamIndex === -1) {
        return {
            name: k,
            params: []
        };
    }
    if (k.at(-1) !== ">") {
        throwParseError(`'>' must be the last character of a generic declaration in a scope`);
    }
    return {
        name: k.slice(0, firstParamIndex),
        params: parseGenericParams(k.slice(firstParamIndex + 1, -1))
    };
};
//# sourceMappingURL=scope.js.map